@@ -5,6 +5,10 @@ from bropkg._util import (
     make_dir,
     find_program,
     read_bro_config_line,
+    std_encoding,
+)
+from bropkg.package import (
+    TRACKING_METHOD_VERSION,
 )
 
 import os
@@ -25,6 +29,13 @@ try:
 except ImportError as err:
     import configparser
 
+if ( sys.version_info[0] < 3 or
+     (sys.version_info[0] == 3 and sys.version_info[1] < 2) ):
+    from configparser import SafeConfigParser as GoodConfigParser
+else:
+    # SafeConfigParser renamed to ConfigParser in Python >= 3.2
+    from configparser import ConfigParser as GoodConfigParser
+
 import git
 
 import bropkg
@@ -32,8 +43,8 @@ import bropkg
 
 def get_input(prompt):
     if sys.version_info[0] < 3:
-        prompt_bytes = prompt.encode(sys.stdout.encoding)
-        return raw_input(prompt_bytes).decode(sys.stdin.encoding)
+        prompt_bytes = prompt.encode(std_encoding(sys.stdout))
+        return raw_input(prompt_bytes).decode(std_encoding(sys.stdin))
 
     return input(prompt)
 
@@ -116,7 +127,7 @@ def prompt_for_user_vars(manager, config, force, pkg_infos):
 
         config_file = os.path.join(config_dir(), 'config')
 
-        with io.open(config_file, 'w', encoding=sys.stdout.encoding) as f:
+        with io.open(config_file, 'w', encoding=std_encoding(sys.stdout)) as f:
             config.write(f)
 
         print('Saved answers to config file: {}'.format(config_file))
@@ -158,7 +169,7 @@ def config_dir():
 
 
 def create_config(configfile):
-    config = configparser.SafeConfigParser()
+    config = GoodConfigParser()
 
     if configfile:
         if not os.path.isfile(configfile):
@@ -226,7 +237,11 @@ def active_git_branch(path):
     if not repo.working_tree_dir:
         return None
 
-    rval = repo.active_branch
+    try:
+        rval = repo.active_branch
+    except TypeError as error:
+        # return detached commit
+        rval = repo.head.commit
 
     if not rval:
         return None
@@ -566,17 +581,18 @@ def cmd_bundle(manager, args, config):
     prefer_existing_clones = False
 
     if args.manifest:
-        manifest_is_file = False
-
-        if len(args.manifest) == 1:
-            config = configparser.SafeConfigParser(delimiters='=')
+        if len(args.manifest) == 1 and os.path.isfile(args.manifest[0]):
+            config = GoodConfigParser(delimiters='=')
             config.optionxform = str
 
             if config.read(args.manifest[0]) and config.has_section('bundle'):
                 packages = config.items('bundle')
-                manifest_is_file = True
+            else:
+                print_error('error: "{}" is not a valid manifest file'.format(
+                    args.manifest[0]))
+                sys.exit(1)
 
-        if not manifest_is_file:
+        else:
             packages = [(name, '') for name in args.manifest]
 
         to_validate = []
@@ -925,7 +941,7 @@ def version_change_string(manager, installed_package):
     new_version = old_version
     version_change = ''
 
-    if installed_package.status.tracking_method == 'version':
+    if installed_package.status.tracking_method == TRACKING_METHOD_VERSION:
         versions = manager.package_versions(installed_package)
 
         if len(versions):
@@ -969,7 +985,8 @@ def cmd_upgrade(manager, args, config):
 
         next_version = ipkg.status.current_version
 
-        if ipkg.status.tracking_method == 'version' and info.versions:
+        if ( ipkg.status.tracking_method == TRACKING_METHOD_VERSION and
+             info.versions ):
             next_version = info.versions[-1]
 
         outdated_packages.append((info, next_version, False))
@@ -1242,7 +1259,7 @@ def cmd_unpin(manager, args, config):
                 'Failed unpinning "{}": no such package installed'.format(name))
 
 
-def cmd_list(manager, args, config):
+def _get_filtered_packages(manager, category):
     pkg_dict = dict()
 
     for ipkg in manager.installed_packages():
@@ -1254,29 +1271,35 @@ def cmd_list(manager, args, config):
         if pkg_qn not in pkg_dict:
             pkg_dict[pkg_qn] = pkg
 
-    if args.category == 'all':
+    if category == 'all':
         filtered_pkgs = pkg_dict
-    elif args.category == 'installed':
+    elif category == 'installed':
         filtered_pkgs = {key: value for key, value in pkg_dict.items()
                          if isinstance(value, bropkg.InstalledPackage)}
-    elif args.category == 'not_installed':
+    elif category == 'not_installed':
         filtered_pkgs = {key: value for key, value in pkg_dict.items()
                          if not isinstance(value, bropkg.InstalledPackage)}
-    elif args.category == 'loaded':
+    elif category == 'loaded':
         filtered_pkgs = {key: value for key, value in pkg_dict.items()
                          if isinstance(value, bropkg.InstalledPackage) and
                          value.status.is_loaded}
-    elif args.category == 'unloaded':
+    elif category == 'unloaded':
         filtered_pkgs = {key: value for key, value in pkg_dict.items()
                          if isinstance(value, bropkg.InstalledPackage) and
                          not value.status.is_loaded}
-    elif args.category == 'outdated':
+    elif category == 'outdated':
         filtered_pkgs = {key: value for key, value in pkg_dict.items()
                          if isinstance(value, bropkg.InstalledPackage) and
                          value.status.is_outdated}
     else:
         raise NotImplementedError
 
+    return filtered_pkgs
+
+
+def cmd_list(manager, args, config):
+    filtered_pkgs = _get_filtered_packages(manager, args.category)
+
     for pkg_name, val in sorted(filtered_pkgs.items()):
         if isinstance(val, bropkg.InstalledPackage):
             pkg = val.package
@@ -1354,8 +1377,19 @@ def cmd_info(manager, args, config):
 
     # Dictionary for storing package info to output as JSON
     pkginfo = dict()
+    had_invalid_package = False
 
-    for name in args.package:
+    if len(args.package) == 1:
+        try:
+            filtered_pkgs = _get_filtered_packages(manager, args.package[0])
+            package_names = [pkg_name
+                             for pkg_name, _ in sorted(filtered_pkgs.items())]
+        except NotImplementedError:
+            package_names = args.package
+    else:
+        package_names = args.package
+
+    for name in package_names:
         info = manager.info(name,
                             version=args.version,
                             prefer_installed=(args.nolocal != True))
@@ -1374,6 +1408,8 @@ def cmd_info(manager, args, config):
                 pkginfo[name]['invalid'] = info.invalid_reason
             else:
                 print('\tinvalid package: {}\n'.format(info.invalid_reason))
+
+            had_invalid_package = True
             continue
 
         if args.json:
@@ -1434,6 +1470,9 @@ def cmd_info(manager, args, config):
     if args.json:
         print(json.dumps(pkginfo, indent=args.jsonpretty, sort_keys=True))
 
+    if had_invalid_package:
+        sys.exit(1)
+
 
 def _fill_metadata_version(pkginfo_name_metadata_version, info_metadata):
     """ Fill a dict with metadata information.
@@ -1531,7 +1570,7 @@ def cmd_autoconfig(manager, args, config):
     change_config_value(config, 'paths', 'bro_dist',
                         bro_dist, config_file_exists)
 
-    with io.open(config_file, 'w', encoding=sys.stdout.encoding) as f:
+    with io.open(config_file, 'w', encoding=std_encoding(sys.stdout)) as f:
         config.write(f)
 
     print('Successfully wrote config file to {}'.format(config_file))
@@ -1623,8 +1662,8 @@ def argparser():
     sub_parser.add_argument(
         '--version', default=None,
         help='The version of the package to test.  Only one package may be'
-        ' specified at a time when using this flag.  A version tag or branch'
-        ' name may be specified here.'
+        ' specified at a time when using this flag.  A version tag, branch'
+        ' name, or commit hash may be specified here.'
         ' If the package name refers to a local git repo with a working tree,'
         ' then its currently active branch is used.'
         ' The default for other cases is to use'
@@ -1659,8 +1698,8 @@ def argparser():
     sub_parser.add_argument(
         '--version', default=None,
         help='The version of the package to install.  Only one package may be'
-        ' specified at a time when using this flag.  A version tag or branch'
-        ' name may be specified here.'
+        ' specified at a time when using this flag.  A version tag, branch'
+        ' name, or commit hash may be specified here.'
         ' If the package name refers to a local git repo with a working tree,'
         ' then its currently active branch is used.'
         ' The default for other cases is to use'
@@ -1673,7 +1712,7 @@ def argparser():
         help='Creates a bundle file containing a collection of Bro packages.',
         description='This command creates a bundle file containing a collection'
                     ' of Bro packages.  If ``--manifest`` is used, the user'
-                    ' suplies the list of packages to put in the bundle, else'
+                    ' supplies the list of packages to put in the bundle, else'
                     ' all currently installed packages are put in the bundle.'
                     ' A bundle file can be unpacked on any target system,'
                     ' resulting in a repeatable/specific set of packages'
@@ -1707,8 +1746,9 @@ def argparser():
         ' in the bundle.  If a file name is supplied, it should be in INI'
         ' format with a single ``[bundle]`` section.  The keys in that section'
         ' correspond to package names and their values correspond to git'
-        ' version tags or branch names.  The values may be left blank to'
-        ' indicate that the latest available version should be used.')
+        ' version tags, branch names, or commit hashes.  The values may be'
+        ' left blank to indicate that the latest available version should be'
+        ' used.')
 
     # unbundle
     sub_parser = command_parser.add_parser(
@@ -1898,7 +1938,10 @@ def argparser():
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
     sub_parser.set_defaults(run_cmd=cmd_info)
     sub_parser.add_argument(
-        'package', nargs='+', default=[], help=pkg_name_help)
+        'package', nargs='+', default=[], help=pkg_name_help +
+        ' If a single name is given and matches one of the same categories'
+        ' as the "list" command, then it is automatically expanded to be the'
+        ' names of all packages which match the given category.')
     sub_parser.add_argument(
         '--version', default=None,
         help='The version of the package metadata to inspect.  A version tag,'
