@@ -4,6 +4,7 @@ to interact with and operate on Bro packages.
 """
 
 import os
+import sys
 import copy
 import json
 import shutil
@@ -16,6 +17,13 @@ try:
 except ImportError as err:
     import configparser
 
+if ( sys.version_info[0] < 3 or
+     (sys.version_info[0] == 3 and sys.version_info[1] < 2) ):
+    from configparser import SafeConfigParser as GoodConfigParser
+else:
+    # SafeConfigParser renamed to ConfigParser in Python >= 3.2
+    from configparser import ConfigParser as GoodConfigParser
+
 import git
 import semantic_version as semver
 
@@ -26,7 +34,7 @@ from ._util import (
     copy_over_path,
     git_clone_shallow,
     get_bro_version,
-    stdout_encoding,
+    std_encoding,
     find_program,
     read_bro_config_line,
 )
@@ -36,6 +44,9 @@ from .source import (
 )
 from .package import (
     METADATA_FILENAME,
+    TRACKING_METHOD_VERSION,
+    TRACKING_METHOD_BRANCH,
+    TRACKING_METHOD_COMMIT,
     name_from_path,
     user_vars,
     canonical_url,
@@ -247,13 +258,12 @@ class Manager(object):
             for dicts in pkg_list:
                 pkg_dict = dicts['package_dict']
                 status_dict = dicts['status_dict']
-
                 pkg_name = pkg_dict['name']
-                del pkg_dict['name']
 
                 if version == 0 and 'index_data' in pkg_dict:
                     del pkg_dict['index_data']
 
+                pkg_dict['canonical'] = True;
                 pkg = Package(**pkg_dict)
                 status = PackageStatus(**status_dict)
                 self.installed_pkgs[pkg_name] = InstalledPackage(pkg, status)
@@ -632,7 +642,8 @@ class Manager(object):
                 LOG.debug("no aggegrate file in source, restore local one")
 
         if aggregate:
-            parser = configparser.SafeConfigParser()
+            # Use raw parser so no value interpolation takes place.
+            parser = configparser.RawConfigParser()
 
             for index_file in source.package_index_files():
                 urls = []
@@ -763,12 +774,16 @@ class Manager(object):
         clonepath = os.path.join(self.package_clonedir, ipkg.package.name)
         clone = git.Repo(clonepath)
 
-        if ipkg.status.tracking_method == 'version':
+        if ipkg.status.tracking_method == TRACKING_METHOD_VERSION:
             version_tags = _get_version_tags(clone)
             return self._install(ipkg.package, version_tags[-1])
-        elif ipkg.status.tracking_method == 'branch':
+        elif ipkg.status.tracking_method == TRACKING_METHOD_BRANCH:
             clone.remote().pull()
             return self._install(ipkg.package, ipkg.status.current_version)
+        elif ipkg.status.tracking_method == TRACKING_METHOD_COMMIT:
+            # The above check for whether the installed package is outdated
+            # also should have already caught this situation.
+            return "package is not outdated"
         else:
             raise NotImplementedError
 
@@ -992,7 +1007,7 @@ class Manager(object):
             return (str(error), infos)
 
         manifest_file = os.path.join(bundle_dir, 'manifest.txt')
-        config = configparser.SafeConfigParser(delimiters='=')
+        config = GoodConfigParser(delimiters='=')
         config.optionxform = str
 
         if not config.read(manifest_file):
@@ -1005,7 +1020,8 @@ class Manager(object):
         manifest = config.items('bundle')
 
         for git_url, version in manifest:
-            package = Package(git_url=git_url)
+            package = Package(git_url=git_url, name=git_url.split('/')[-1],
+                              canonical=True)
             pkg_path = os.path.join(bundle_dir, package.name)
             pkg_info = self.info(pkg_path, version=version,
                                  prefer_installed=False)
@@ -1063,7 +1079,9 @@ class Manager(object):
 
             LOG.info('getting info on "%s": matched no source package',
                      pkg_path)
-            reason = 'package name not found in sources and not a valid git URL'
+            reason = ('package name not found in sources and also'
+                      ' not a usable git URL (invalid or inaccessible,'
+                      ' use -vvv for details)')
             return PackageInfo(package=package, invalid_reason=reason,
                                status=status)
 
@@ -1206,7 +1224,7 @@ class Manager(object):
 
             node = Node(info.package.qualified_name())
             node.info = info
-            method = 'version' if version in node.info.versions else 'branch'
+            method = node.info.version_type
             node.requested_version = (method, version)
             graph[node.name] = node
 
@@ -1275,14 +1293,14 @@ class Manager(object):
 
             if bro_version:
                 node = Node('bro')
-                node.installed_version = ('version', bro_version)
+                node.installed_version = (TRACKING_METHOD_VERSION, bro_version)
                 graph['bro'] = node
             else:
                 LOG.warning(
                     'could not get bro version: no bro_config in PATH?')
 
             node = Node('bro-pkg')
-            node.installed_version = ('version', __version__)
+            node.installed_version = (TRACKING_METHOD_VERSION, __version__)
             graph['bro-pkg'] = node
 
             for ipkg in self.installed_packages():
@@ -1358,7 +1376,7 @@ class Manager(object):
                 # Check that requested version doesn't conflict with dependers.
                 track_method, required_version = node.requested_version
 
-                if track_method == 'branch':
+                if track_method == TRACKING_METHOD_BRANCH:
                     for depender_name, version_spec in node.dependers.items():
                         if version_spec == '*':
                             continue
@@ -1369,13 +1387,28 @@ class Manager(object):
                         if version_spec == required_version:
                             continue
 
+                        return (str.format(
+                            'unsatisfiable dependency: requested "{}" ({}),'
+                            ' but "{}" requires {}', node.name,
+                            required_version, depender_name, version_spec),
+                            new_pkgs)
+                elif track_method == TRACKING_METHOD_COMMIT:
+                    for depender_name, version_spec in node.dependers.items():
+                        if version_spec == '*':
+                            continue
+
+                        # Could allow commit= version specification like what
+                        # is done with branches, but unsure there's a common
+                        # use-case for it.
+
                         return (str.format(
                             'unsatisfiable dependency: requested "{}" ({}),'
                             ' but "{}" requires {}', node.name,
                             required_version, depender_name, version_spec),
                             new_pkgs)
                 else:
-                    req_semver = semver.Version.coerce(required_version)
+                    normal_version = _normalize_version_tag(required_version)
+                    req_semver = semver.Version.coerce(normal_version)
 
                     for depender_name, version_spec in node.dependers.items():
                         if version_spec.startswith('branch='):
@@ -1403,7 +1436,7 @@ class Manager(object):
                 # Check that installed version doesn't conflict with dependers.
                 track_method, required_version = node.installed_version
 
-                if track_method == 'branch':
+                if track_method == TRACKING_METHOD_BRANCH:
                     for depender_name, version_spec in node.dependers.items():
                         if version_spec == '*':
                             continue
@@ -1414,13 +1447,28 @@ class Manager(object):
                         if version_spec == required_version:
                             continue
 
+                        return (str.format(
+                            'unsatisfiable dependency: "{}" ({}) is installed,'
+                            ' but "{}" requires {}', node.name,
+                            required_version, depender_name, version_spec),
+                            new_pkgs)
+                elif track_method == TRACKING_METHOD_COMMIT:
+                    for depender_name, version_spec in node.dependers.items():
+                        if version_spec == '*':
+                            continue
+
+                        # Could allow commit= version specification like what
+                        # is done with branches, but unsure there's a common
+                        # use-case for it.
+
                         return (str.format(
                             'unsatisfiable dependency: "{}" ({}) is installed,'
                             ' but "{}" requires {}', node.name,
                             required_version, depender_name, version_spec),
                             new_pkgs)
                 else:
-                    req_semver = semver.Version.coerce(required_version)
+                    normal_version = _normalize_version_tag(required_version)
+                    req_semver = semver.Version.coerce(normal_version)
 
                     for depender_name, version_spec in node.dependers.items():
                         if version_spec.startswith('branch='):
@@ -1507,7 +1555,8 @@ class Manager(object):
                             best_version = 'master'
                     elif need_version:
                         for version in node.info.versions[::-1]:
-                            req_semver = semver.Version.coerce(version)
+                            normal_version = _normalize_version_tag(version)
+                            req_semver = semver.Version.coerce(normal_version)
 
                             satisfied = True
 
@@ -1560,7 +1609,7 @@ class Manager(object):
         delete_path(bundle_dir)
         make_dir(bundle_dir)
         manifest_file = os.path.join(bundle_dir, 'manifest.txt')
-        config = configparser.SafeConfigParser(delimiters='=')
+        config = GoodConfigParser(delimiters='=')
         config.optionxform = str
         config.add_section('bundle')
 
@@ -1632,7 +1681,7 @@ class Manager(object):
             return str(error)
 
         manifest_file = os.path.join(bundle_dir, 'manifest.txt')
-        config = configparser.SafeConfigParser(delimiters='=')
+        config = GoodConfigParser(delimiters='=')
         config.optionxform = str
 
         if not config.read(manifest_file):
@@ -1644,7 +1693,8 @@ class Manager(object):
         manifest = config.items('bundle')
 
         for git_url, version in manifest:
-            package = Package(git_url=git_url)
+            package = Package(git_url=git_url, name=git_url.split('/')[-1],
+                              canonical=True)
             clonepath = os.path.join(self.package_clonedir, package.name)
             delete_path(clonepath)
             shutil.move(os.path.join(bundle_dir, package.name), clonepath)
@@ -1811,8 +1861,7 @@ class Manager(object):
             if k not in substitutions:
                 substitutions[k] = v
 
-        metadata_parser = configparser.SafeConfigParser(
-            defaults=substitutions)
+        metadata_parser = GoodConfigParser(defaults=substitutions)
         invalid_reason = _parse_package_metadata(
             metadata_parser, metadata_file)
 
@@ -1840,7 +1889,7 @@ class Manager(object):
                     LOG.info('installing "%s": writing build log: %s',
                              package, buildlog)
 
-                    f.write(u'=== STDERR ===\n'.encode(stdout_encoding()))
+                    f.write(u'=== STDERR ===\n'.encode(std_encoding(sys.stderr)))
 
                     while True:
                         data = build.stderr.read(bufsize)
@@ -1850,7 +1899,7 @@ class Manager(object):
                         else:
                             break
 
-                    f.write(u'=== STDOUT ===\n'.encode(stdout_encoding()))
+                    f.write(u'=== STDOUT ===\n'.encode(std_encoding(sys.stdout)))
 
                     while True:
                         data = build.stdout.read(bufsize)
@@ -1939,8 +1988,8 @@ class Manager(object):
 
             version (str): if not given, then the latest git version tag is
                 installed (or if no version tags exist, the "master" branch is
-                installed).  If given, it may be either a git version tag or a
-                git branch name.
+                installed).  If given, it may be either a git version tag, a
+                git branch name, or a git commit hash.
 
         Returns:
             str: empty string if package installation succeeded else an error
@@ -2026,13 +2075,15 @@ class Manager(object):
         version_tags = _get_version_tags(clone)
 
         if version:
-            if version in version_tags:
-                status.tracking_method = 'version'
+            if _is_commit_hash(clone, version):
+                status.tracking_method = TRACKING_METHOD_COMMIT
+            elif version in version_tags:
+                status.tracking_method = TRACKING_METHOD_VERSION
             else:
                 branches = _get_branch_names(clone)
 
                 if version in branches:
-                    status.tracking_method = 'branch'
+                    status.tracking_method = TRACKING_METHOD_BRANCH
                 else:
                     LOG.info(
                         'branch "%s" not in available branches: %s', version,
@@ -2042,16 +2093,16 @@ class Manager(object):
         else:
             if len(version_tags):
                 version = version_tags[-1]
-                status.tracking_method = 'version'
+                status.tracking_method = TRACKING_METHOD_VERSION
             else:
                 if 'master' not in _get_branch_names(clone):
                     return 'git repo has no "master" branch or version tags'
 
                 version = 'master'
-                status.tracking_method = 'branch'
+                status.tracking_method = TRACKING_METHOD_BRANCH
 
         status.current_version = version
-        status.current_hash = _get_hash(clone, version)
+        status.current_hash = _get_hash(clone, version, status.tracking_method)
         clone.git.checkout(version)
         status.is_outdated = _is_clone_outdated(
             clone, version, status.tracking_method)
@@ -2089,22 +2140,29 @@ class Manager(object):
         LOG.debug('installed "%s"', package)
         return ''
 
+def _normalize_version_tag(tag):
+    # Change vX.Y.Z into X.Y.Z
+    if len(tag) > 1 and tag[0] is 'v' and tag[1].isdigit():
+        return tag[1:]
+
+    return tag
 
 def _get_version_tags(clone):
     tags = []
 
     for tagref in clone.tags:
         tag = str(tagref.name)
+        normal_tag = _normalize_version_tag(tag)
 
         try:
-            semver.Version.coerce(tag)
+            sv = semver.Version.coerce(normal_tag)
         except ValueError:
             # Skip tags that aren't compatible semantic versions.
             continue
         else:
-            tags.append(tag)
+            tags.append((normal_tag, tag, sv))
 
-    return sorted(tags, key=semver.Version.coerce)
+    return [t[1] for t in sorted(tags, key=lambda e: e[2])]
 
 
 def _get_branch_names(clone):
@@ -2116,20 +2174,24 @@ def _get_branch_names(clone):
         if not branch_name.startswith('origin/'):
             continue
 
-        rval.append(branch_name.split('/')[-1])
+        rval.append(branch_name.split('origin/')[1])
 
     return rval
 
 
-def _get_ref(clone, ref_name):
+def _get_ref(clone, ref_name, track_method):
     for ref in clone.refs:
-        if ref.name.split('/')[-1] == ref_name:
+        if ((track_method == TRACKING_METHOD_BRANCH and
+             'origin' in ref.name and
+             ref.name.split('origin/')[1] == ref_name) or
+             ref.name.split('/')[-1] == ref_name):
             return ref
 
 
 def _is_version_outdated(clone, version):
     version_tags = _get_version_tags(clone)
-    return version != version_tags[-1]
+    latest = _normalize_version_tag(version_tags[-1])
+    return _normalize_version_tag(version) != latest
 
 
 def _is_branch_outdated(clone, branch):
@@ -2139,16 +2201,29 @@ def _is_branch_outdated(clone, branch):
 
 
 def _is_clone_outdated(clone, ref_name, tracking_method):
-    if tracking_method == 'version':
+    if tracking_method == TRACKING_METHOD_VERSION:
         return _is_version_outdated(clone, ref_name)
-    elif tracking_method == 'branch':
+    elif tracking_method == TRACKING_METHOD_BRANCH:
         return _is_branch_outdated(clone, ref_name)
+    elif tracking_method == TRACKING_METHOD_COMMIT:
+        return False
     else:
         raise NotImplementedError
 
 
-def _get_hash(clone, ref_name):
-    return _get_ref(clone, ref_name).object.hexsha
+def _get_hash(clone, version, track_method):
+    if track_method == TRACKING_METHOD_COMMIT:
+        return clone.commit(version).hexsha
+
+    return _get_ref(clone, version, track_method).object.hexsha
+
+
+def _is_commit_hash(clone, text):
+    try:
+        commit = clone.commit(text)
+        return commit.hexsha.startswith(text)
+    except Exception:
+        return False
 
 
 def _copy_package_dir(package, dirname, src, dst, scratch_dir):
@@ -2256,6 +2331,14 @@ def _info_from_clone(clone, package, status, version):
         A :class:`.package.PackageInfo` object.
     """
     versions = _get_version_tags(clone)
+
+    if _is_commit_hash(clone, version):
+        version_type = TRACKING_METHOD_COMMIT
+    elif version in versions:
+        version_type = TRACKING_METHOD_VERSION
+    else:
+        version_type = TRACKING_METHOD_BRANCH
+
     metadata_file = os.path.join(clone.working_dir, METADATA_FILENAME)
     # Use raw parser so no value interpolation takes place.
     metadata_parser = configparser.RawConfigParser()
@@ -2265,10 +2348,10 @@ def _info_from_clone(clone, package, status, version):
     if invalid_reason:
         return PackageInfo(package=package, invalid_reason=invalid_reason,
                            status=status, versions=versions,
-                           metadata_version=version)
+                           metadata_version=version, version_type=version_type)
 
     metadata = _get_package_metadata(metadata_parser)
 
     return PackageInfo(package=package, invalid_reason=invalid_reason,
                        status=status, metadata=metadata, versions=versions,
-                       metadata_version=version)
+                       metadata_version=version, version_type=version_type)
